/**
 * Constants
 */
var gc = gc || {};

var COMMAND_PACKET = 1;
var REPLY_PACKET = 2;
var ERROR_PACKET = 3;
var PAYLOAD_PACKET = 4;
var INTERRUPT_PACKET = 5;

// packet flags
var FLAG_MOREDATA = 1;

var PACKET_IDENTIFIER = 'T'.charCodeAt(0);

var PACKET_ID = 0;
var PACKET_PEC = 1;
var PACKET_PAYLOAD_LEN = 2;
var PACKET_TYPE = 3;
var PACKET_FLAGS = 4;
var PACKET_SEQ_NUM = 5;
var PACKET_STATUS = 6;
var PACKET_COMMAND = 7;
var PACKET_PAYLOAD = 8;

var MAX_PACKET_SIZE = 62; // size, in bytes, of a USB packet
var PACKET_HEADER_SIZE = PACKET_PAYLOAD;
var MAX_PAYLOAD = MAX_PACKET_SIZE - PACKET_HEADER_SIZE;

var Command =
    {
      Cmd_I2C_Control : 1, // 0x01
      Cmd_I2C_BlkWriteBlkRead : 96, // 0x60
    };

var init = function() {

  /***
   * I2C Interface sub-module definition
   ***/
  var I2C_100kHz = 0;
  var I2C_400kHz = 1;
  var I2C_10kHz  = 2;
  var I2C_800kHz = 3;
  
  var I2C_7Bits = 0;
  var I2C_10Bits = 1;

  var I2C_PullUps_OFF = 0;
  var I2C_PullUps_ON = 1;
  
  var CustomI2C_Interface = function(settings, registerModel, u2a) {
    // console.log('Custom I2C', 'Constructor', settings, registerModel, u2a);
    this._registerModel = registerModel;
    this.readyToProcess = true;
    this.cmdQueue = [];
  };

  /**
   * Add the custom interface to the USB2ANY prototype, the name must take the form of "XXX_Interface" where "XXX" is all capitalized, eg. "LED_Interface"
   * To use custom interfaces in the system.json, set the name property to be the "XXX" part of the interface name (not case-sensitive), eg. "name" : "led"
   * To avoid naming conflicts with already defined interfaces, do not use names "power", "i2c", "spi", "gpio", "uart", "pwm" or "displayscale"
   */
  gc.databind.USB2ANY.prototype.CUSTOMI2C_Interface = CustomI2C_Interface;

  /**
   * Custom interfaces should be extended from the ICommInterface base class
   */
  CustomI2C_Interface.prototype = new gc.databind.USB2ANY.ICommInterface();
  
  CustomI2C_Interface.prototype.control = function(u2a, settings) {
    // console.log('Custom I2C', 'Control', u2a, settings);

    var speed = I2C_10kHz;
    switch(settings.speed)
    {
      case 100:
        speed = I2C_100kHz;
        break;
      case 400:
        speed = I2C_400kHz;
        break;
      case 800:
        speed = I2C_800kHz;
        break;
    }
    
    var addrsBits = I2C_7Bits;
    switch(settings.addrsBits)
    {
      case 10:
        addrsBits = I2C_10Bits;
        break;
    }
    
    var pullUps = settings.pullup ? I2C_PullUps_ON : I2C_PullUps_OFF;
    
    u2a.sendCommandPacket(Command.Cmd_I2C_Control,
    [
      speed, addrsBits, pullUps
    ]);
  };

  CustomI2C_Interface.prototype.initSymbolsForDevice = function(settings, registerModel)
  {
      registerModel.readDeviceAddrsMap('I2C', settings);
  };

  CustomI2C_Interface.prototype.manageReadWrite = function(u2a, cmdInfo) {
    if(cmdInfo) {
      this.cmdQueue.push(cmdInfo);
    }

    if(this.readyToProcess && this.cmdQueue.length > 0) {
      var cmd = this.cmdQueue.shift();

      switch(cmd.type) {
        case 'read':
          this.readyToProcess = false;

          var data = [];
          data.push.apply(data, cmd.i2cAddr);
          data.push(cmd.regAddrSize);
          data.push.apply(data, cmd.noOfReg);
          data.push(cmd.regAddr);
          u2a.readResponse(u2a.sendCommandPacket(101, data)).then(function(packet) {
            // console.log('Custom I2C', 'Read', 'Response', packet);
            cmd.promise.resolve(packet);

            this.readyToProcess = true;
            this.manageReadWrite(u2a);
          }.bind(this));
          break;

        case 'write':
          this.readyToProcess = false;

          var data = [];
          data.push.apply(data, cmd.i2cAddr);
          data.push(cmd.dataLen);
          data.push(cmd.regAddr);
          data.push.apply(data, cmd.value);
          u2a.readResponse(u2a.sendCommandPacket(2, data)).then(function(packet) {
            // console.log('Custom I2C', 'Write', 'Response', packet);
            cmd.promise.resolve(packet);

            this.readyToProcess = true;
            this.manageReadWrite(u2a);
          }.bind(this));
          break;
      }
    }
  };

  CustomI2C_Interface.prototype.read = function(u2a, info, registerModel, coreIndex) {
    // console.log('Custom I2C', 'Read', u2a, info, registerModel, coreIndex);
    // console.log('Custom I2C', 'Read', 'Device Address', info.deviceAddrs, this._registerModel.getDeviceAddrsForRegister(info));

    var deferred = Q.defer(), readPromise = Q.defer();

    var i2cAddrArr = [0x00, 0x00];
    var devAddrFromRegModel = parseInt(this._registerModel.getDeviceAddrsForRegister(info));
    if(!isNaN(devAddrFromRegModel)) {
      i2cAddrArr[0] = (devAddrFromRegModel >> 8) & 0xFF;
      i2cAddrArr[1] = devAddrFromRegModel & 0xFF;
    }

    var noOfRegArr = [], nBytes = info.nBytes ? info.nBytes : 1;
    noOfRegArr[0] = (nBytes >> 8) & 0xFF;
    noOfRegArr[1] = nBytes & 0xFF;

    this.manageReadWrite(u2a, {
      type: 'read',
      promise: deferred,
      i2cAddr: i2cAddrArr,
      regAddrSize: 1,
      noOfReg: noOfRegArr,
      regAddr: info.addr
    });

    deferred.promise.then(function(packet) {
      var value = 0, valueArr = packet.splice(8, info.nBytes);
      for(var i=0; i<valueArr.length; i++) {
        value = (value << 8) | valueArr[i];
      }
      readPromise.resolve(value);
    });

    return readPromise.promise;
  };
  
  CustomI2C_Interface.prototype.write = function(u2a, info, value, registerModel, coreIndex) {
    // console.log('Custom I2C', 'Write', u2a, info, value, registerModel, coreIndex);
    // console.log('Custom I2C', 'Write', 'Device Address', info.deviceAddrs, this._registerModel.getDeviceAddrsForRegister(info));

    var deferred = Q.defer(), writePromise = Q.defer();

    var i2cAddrArr = [0x00, 0x00];
    var devAddrFromRegModel = parseInt(this._registerModel.getDeviceAddrsForRegister(info));
    if(!isNaN(devAddrFromRegModel)) {
      i2cAddrArr[0] = (devAddrFromRegModel >> 8) & 0xFF;
      i2cAddrArr[1] = devAddrFromRegModel & 0xFF;
    }

    var valueArr = [];
    for(var i=0; i<info.nBytes; i++) {
      noOfShifts = 8 * (info.nBytes - i -1);
      valueArr[i] = (value & (0xFF << noOfShifts)) >> noOfShifts;
    }

    this.manageReadWrite(u2a, {
      type: 'write',
      promise: deferred,
      i2cAddr: i2cAddrArr,
      dataLen: info.nBytes + 1,
      regAddr: info.addr,
      value: valueArr
    });

    deferred.promise.then(function(packet) {
      writePromise.resolve(true);
    });

    return writePromise.promise;
  };

};


// Wait until USB2ANY global var is ready before defining interface
// if (gc.databind.USB2ANY) {
//     init();
// } else {
//     document.addEventListener('DOMContentLoaded', init);
// }
var templateObj = document.querySelector('#template_obj');
if (templateObj) {
    init();
} else {
    if(gc.desktop.isNodeWebkit()) {
      document.addEventListener('DOMContentLoaded', init);
    } else {
      document.addEventListener('WebComponentsReady', init);
    }
}